/***********************************************************************
*
* mimedefang-multiplexor.c
*
* Master program which manages a pool of e-mail scanning processes.
*
* Copyright (C) 2001-2003 Roaring Penguin Software Inc.
* http://www.roaringpenguin.com
*
* This program may be distributed under the terms of the GNU General
* Public License, Version 2, or (at your option) any later version.
*
***********************************************************************/

static char const RCSID[] =
"$Id: mimedefang-multiplexor.c,v 1.128 2003/04/21 15:53:14 dfs Exp $";

#include "config.h"
#include "event_tcp.h"

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <stdarg.h>
#include <pwd.h>

#ifdef HAVE_SETRLIMIT
#include <sys/resource.h>
static void limit_mem_usage(unsigned long rss, unsigned long as);
#endif

/* Weird case, but hey... */
#if defined(HAVE_WAIT3) && !defined(HAVE_SETRLIMIT)
#include <sys/resource.h>
#endif

#ifndef AF_LOCAL
#define AF_LOCAL AF_UNIX
#endif

#ifndef SUN_LEN
#define SUN_LEN(ptr)  ((size_t) (((struct sockaddr_un *) 0)->sun_path) \
       + strlen ((ptr)->sun_path))
#endif

#define MAX_CMD_LEN 4096	/* Maximum length of command from mimedefang */
#define MAX_DIR_LEN  512        /* Maximum lenght of working directory */

#define DOLOG Settings.doSyslog

#define SLAVENO(s) ((int) ((s) - AllSlaves))

/* A slave can be in one of four states:
   Stopped -- Slave has no associated Perl process
   Idle    -- Slave has an associated process, but is not doing work
   Busy    -- Slave is processing a command
   Killed  -- Slave has been killed, but we're waiting for it to exit */

#define STATE_STOPPED    0
#define STATE_IDLE       1
#define STATE_BUSY       2
#define STATE_KILLED     3
#define NUM_SLAVE_STATES 4

/* Structure of a slave process */
typedef struct Slave_t {
    struct Slave_t *next;	/* Link in free/busy list                   */
    EventSelector *es;		/* Event selector                           */
    pid_t pid;			/* Process ID of slave process              */
    int numRequests;		/* Number of requests handled by slave      */
    int numScans;               /* Number of messages scanned               */
    time_t idleTime;		/* Time when slave became idle              */
    unsigned int activated;     /* Activation order                         */
    int slaveStdin;		/* Slave's stdin descriptor                 */
    int slaveStdout;		/* Slave's stdout descriptor                */
    int slaveStderr;		/* Slave's stderr descriptor                */
    int clientFD;		/* Client file descriptor                   */
    int oom;			/* Did slave run out of memory?             */
    EventTcpState *event;	/* Pending event handler                    */
    EventHandler *errHandler;	/* Read handler for stderr                  */
    EventHandler *termHandler;  /* Timer after which we send SIGTERM        */
    char workdir[MAX_DIR_LEN];	/* Working directory for current scan       */
    int generation;		/* Slave's generation                       */
    int state;			/* Slave's state                            */
} Slave;

Slave *AllSlaves;		/* Array of all slaves                      */
Slave *Slaves[NUM_SLAVE_STATES]; /* Lists of slaves in each state           */
int SlaveCount[NUM_SLAVE_STATES]; /* Count of slaves in each state          */
int Generation = 0;             /* Current generation                       */
int NumMsgsProcessed = 0;	/* Number of messages processed since last
				   "msgs" query */
unsigned int Activations = 0;	/* Incremented when a slave is activated    */

struct Settings_t {
    int minSlaves;		/* Minimum number of slaves to keep running */
    int maxSlaves;		/* Maximum possible number of slaves        */
    int maxRequests;		/* Maximum number of requests per slave     */
    int maxIdleTime;		/* Excess slaves should be killed after time */
    int busyTimeout;		/* Timeout after which we kill scanner      */
    int clientTimeout;		/* Timeout for client request/reply         */
    int slewTime;		/* Time to wait between slaves' activation  */
    int waitTime;               /* Minimum time to wait between activations */
    int doSyslog;		/* If true, log various things with syslog  */
    char const *sockName;	/* Socket name for talking to mimedefang    */
    char const *progPath;	/* Program to execute for filter            */
    char const *statsFile;	/* File name for logging statistics         */
    char const *subFilter;      /* Sub-filter to pass to filter             */
    FILE *statsFP;		/* File pointer for stats file              */
    int statsToSyslog;		/* If true, log stats using syslog          */
    int flushStats;             /* If non-zero, flush stats file after write*/
    unsigned long maxRSS;	/* Maximum RSS for slaves (if supported)    */
    unsigned long maxAS;        /* Maximum address space for slaves         */
    int logStatusInterval;      /* How often to log status to syslog        */
} Settings;

/* Pipe written on reception of SIGCHLD */
static int Pipe[2] = {-1, -1};

#ifndef HAVE_SIG_ATOMIC_T
#define sig_atomic_t int
#endif

static volatile sig_atomic_t ReapPending = 0;
static volatile sig_atomic_t HupPending = 0;
static volatile sig_atomic_t IntPending = 0;
static volatile sig_atomic_t CharPending = 0;

static int DebugEvents = 0;
static time_t LastSlaveActivation = (time_t) 0;

extern int drop_privs(char const *user, uid_t uid, gid_t gid);
extern int find_syslog_facility(char const *facility_name);

/* Prototypes */
#ifdef HAVE_WAIT3
static void log_slave_resource_usage(Slave *s, struct rusage *usage);
#endif

static Slave *findFreeSlave(void);
static void shutDescriptors(Slave *s);
static void reapTerminatedSlaves(void);
static Slave *findSlaveByPid(pid_t pid);

static pid_t activateSlave(Slave *s, char const *reason);
static void killSlave(Slave *s, char const *reason);
static void terminateSlave(EventSelector *es, int fd, unsigned int flags,
			   void *data);
static void nukeSlave(EventSelector *es, int fd, unsigned int flags,
		      void *data);

/* List-management functions */
static void unlinkFromList(Slave *s);
static void putOnList(Slave *s, int state);

static void handleAccept(EventSelector *es, int fd);
static void handleCommand(EventSelector *es, int fd,
			  char *buf, int len, int flag, void *data);
static void handleSlaveReceivedCommand(EventSelector *es, int fd,
				       char *buf, int len, int flag,
				       void *data);
static void handleSlaveReceivedAnswer(EventSelector *es, int fd,
				      char *buf, int len, int flag,
				      void *data);
static void doScan(EventSelector *es, int fd, char *cmd);
static void doStatus(EventSelector *es, int fd);

static void doSlaveCommand(EventSelector *es, int fd, char *cmd);
static void checkSlaveForExpiry(Slave *s);
static void handlePipe(EventSelector *es,
		       int fd, unsigned int flags, void *data);
static void handleSlaveStderr(EventSelector *es,
			      int fd,
			      unsigned int flags,
			      void *data);
static void childHandler(int sig);
static void hupHandler(int sig);
static void intHandler(int sig);
static void sigterm(int sig);
static void newGeneration(void);

static void handleIdleTimeout(EventSelector *es, int fd, unsigned int flags,
			      void *data);
static void doStatusLog(EventSelector *es, int fd, unsigned int flags,
			void *data);

static void logSlaveReaped(Slave *s, int status);

static void statsReopenFile(void);
static void statsLog(char const *event, int slaveno, char const *fmt, ...);
static void bringSlavesUpToMin(EventSelector *es, int fd, unsigned int flags,
			       void *data);
static void scheduleBringSlavesUpToMin(EventSelector *es);
static int minScheduled = 0;

#define NUM_FREE_SLAVES    (SlaveCount[STATE_IDLE] + SlaveCount[STATE_STOPPED])
#define NUM_RUNNING_SLAVES (SlaveCount[STATE_IDLE] + SlaveCount[STATE_BUSY] + SlaveCount[STATE_KILLED])

/**********************************************************************
* %FUNCTION: state_name
* %ARGUMENTS:
*  s -- a slave
* %RETURNS:
*  A string representing the name of the state
***********************************************************************/
static char const *
state_name(Slave *s)
{
    switch(s->state) {
    case STATE_STOPPED: return "Stopped";
    case STATE_IDLE:    return "Idle";
    case STATE_BUSY:    return "Busy";
    case STATE_KILLED:  return "Killed";
    }
    return "Unknown";
}

/**********************************************************************
* %FUNCTION: state_name_lc
* %ARGUMENTS:
*  s -- a slave
* %RETURNS:
*  A string representing the name of the state in all lower-case
***********************************************************************/
static char const *
state_name_lc(Slave *s)
{
    switch(s->state) {
    case STATE_STOPPED: return "stopped";
    case STATE_IDLE:    return "idle";
    case STATE_BUSY:    return "busy";
    case STATE_KILLED:  return "killed";
    }
    return "unknown";
}

/**********************************************************************
* %FUNCTION: reply_to_mimedefang_with_len
* %ARGUMENTS:
*  es -- event selector
*  fd -- file descriptor
*  msg -- message to send back
*  len -- length of message
* %RETURNS:
*  The event associated with the reply, or NULL.
* %DESCRIPTION:
*  Sends a final message back to mimedefang.
***********************************************************************/
static EventTcpState *
reply_to_mimedefang_with_len(EventSelector *es,
			     int fd,
			     char const *msg,
			     int len)
{
    EventTcpState *e;
    e = EventTcp_WriteBuf(es, fd, msg, len, NULL,
			  Settings.clientTimeout, NULL);
    if (!e) {
	if (DOLOG) {
	    syslog(LOG_ERR, "reply_to_mimedefang: EventTcp_WriteBuf failed: %m");
	}
	close(fd);
    }
    return e;
}

/**********************************************************************
* %FUNCTION: reply_to_mimedefang
* %ARGUMENTS:
*  es -- event selector
*  fd -- file descriptor
*  msg -- message to send back
* %RETURNS:
*  The event associated with the reply, or NULL.
* %DESCRIPTION:
*  Sends a final message back to mimedefang.
***********************************************************************/
static EventTcpState *
reply_to_mimedefang(EventSelector *es,
		    int fd,
		    char const *msg)
{
    return reply_to_mimedefang_with_len(es, fd, msg, strlen(msg));
}

/**********************************************************************
* %FUNCTION: findSlaveByPid
* %ARGUMENTS:
*  pid -- Process-ID we're looking for
* %RETURNS:
*  The slave with given pid, or NULL if not found
* %DESCRIPTION:
*  Searches the killed, idle and busy lists for specified slave.
***********************************************************************/
static Slave *
findSlaveByPid(pid_t pid)
{
    Slave *s;

    /* Most likely to be on killed list, so search there first */
    s = Slaves[STATE_KILLED];
    while(s) {
	if (s->pid == pid) return s;
	s = s->next;
    }

    s = Slaves[STATE_IDLE];
    while(s) {
	if (s->pid == pid) return s;
	s = s->next;
    }

    s = Slaves[STATE_BUSY];
    while(s) {
	if (s->pid == pid) return s;
	s = s->next;
    }

    return NULL;
}

/**********************************************************************
* %FUNCTION: usage
* %ARGUMENTS:
*  None
* %RETURNS:
*  Nothing (exits)
* %DESCRIPTION:
*  Prints usage information
***********************************************************************/
static void
usage(void)
{
    fprintf(stderr, "mimedefang-multiplexor version %s\n", VERSION);
    fprintf(stderr, "Usage: mimedefang-multiplexor [options]\n");
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "  -h                -- Print usage info and exit\n");
    fprintf(stderr, "  -t filename       -- Log statistics to filename\n");
    fprintf(stderr, "  -p filename       -- Write process-ID in filename\n");
    fprintf(stderr, "  -T                -- Log statistics to syslog\n");
    fprintf(stderr, "  -u                -- Flush stats file after each write\n");
    fprintf(stderr, "  -U username       -- Run as username, not root\n");
    fprintf(stderr, "  -m minSlaves      -- Minimum number of slaves\n");
    fprintf(stderr, "  -x maxSlaves      -- Maximum number of slaves\n");
    fprintf(stderr, "  -r maxRequests    -- Maximum number of requests per slave\n");
    fprintf(stderr, "  -i idleTime       -- Idle time (seconds) for killing excess slaves\n");
    fprintf(stderr, "  -b busyTime       -- Busy time (seconds) for killing hung slaves\n");
    fprintf(stderr, "  -c cmdTime        -- Request/reply transmission timeout (seconds)\n");
    fprintf(stderr, "  -w waitTime       -- How long to wait between slave activations (seconds)\n");
    fprintf(stderr, "  -W waitTime       -- Absolute minimum to wait between slave activations\n");
    fprintf(stderr, "  -s sock           -- UNIX-domain socket for communication\n");
    fprintf(stderr, "  -f /dir/filter    -- Specify full path of filter program\n");
    fprintf(stderr, "  -d                -- Debug events in /var/log/mdefang-event-debug.log\n");
    fprintf(stderr, "  -l                -- Log events with syslog\n");
#ifdef HAVE_SETRLIMIT
    fprintf(stderr, "  -R size           -- Limit RSS to size kB (if supported on your OS)\n");
    fprintf(stderr, "  -M size           -- Limit memory address space to size kB\n");
#endif
    fprintf(stderr, "  -L interval       -- Log slave status every interval seconds\n");
    fprintf(stderr, "  -S facility       -- Set syslog(3) facility\n");
    exit(EXIT_FAILURE);
}

/**********************************************************************
* %FUNCTION: main
* %ARGUMENTS:
*  argc, argv -- usual suspects
* %RETURNS:
* Nothing -- runs in an infinite loop
* %DESCRIPTION:
*  Main program
***********************************************************************/
int
main(int argc, char *argv[])
{
    int i;
    int sock;
    int c;
    int n;
    FILE *fp;
    char *pidfile = NULL;
    char *user = NULL;
    char *options;
    int facility = LOG_MAIL;

    struct sockaddr_un addr;
    EventSelector *es;
    struct sigaction act;
    struct timeval t;
    struct passwd *pw = NULL;

    /* Paranoia time */
    umask(077);

    /* Paranoia time II */
    if (getuid() != geteuid()) {
	fprintf(stderr, "ERROR: %s is NOT intended to run suid! Exiting.\n",
		argv[0]);
	exit(1);
    }

    if (getgid() != getegid()) {
	fprintf(stderr, "ERROR: %s is NOT intended to run sgid! Exiting.\n",
		argv[0]);
	exit(1);
    }

    Settings.minSlaves     = 0;
    Settings.maxSlaves     = 2;
    Settings.maxRequests   = 500;
    Settings.maxIdleTime   = 300;
    Settings.busyTimeout   = 120;
    Settings.slewTime      = 3;
    Settings.waitTime      = 0;
    Settings.clientTimeout = 10;
    Settings.doSyslog      = 0;
    Settings.sockName      = SPOOLDIR "/mimedefang-multiplexor.sock";
    Settings.progPath      = MIMEDEFANG_PL;
    Settings.subFilter     = NULL;
    Settings.statsFile     = NULL;
    Settings.statsFP       = NULL;
    Settings.flushStats    = 0;
    Settings.statsToSyslog = 0;
    Settings.maxRSS = 0;
    Settings.maxAS = 0;
    Settings.logStatusInterval = 0;
#ifndef HAVE_SETRLIMIT
    options = "Tt:um:x:r:i:b:c:s:hdlf:p:w:F:W:U:S:";
#else
    options = "Tt:um:x:r:i:b:c:s:hdlf:p:w:F:W:U:R:M:L:S:";
#endif
    while((c = getopt(argc, argv, options)) != -1) {
	switch(c) {
	case 'S':
	    facility = find_syslog_facility(optarg);
	    if (facility < 0) {
		fprintf(stderr, "Unknown syslog facility %s\n", optarg);
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'L':
	    if (sscanf(optarg, "%d", &n) != 1) usage();
	    if (n <= 0) {
		n = 0;
	    } else if (n < 5) {
		n = 5;
	    }
	    Settings.logStatusInterval = n;
	    break;

	case 'R':
	case 'M':
	    if (sscanf(optarg, "%d", &n) != 1) usage();
	    if (c == 'R') {
		Settings.maxRSS = (unsigned long) n;
	    } else {
		Settings.maxAS = (unsigned long) n;
	    }
	    break;

	case 'U':
	    /* User to run as */
	    if (user) {
		free(user);
	    }
	    user = strdup(optarg);
	    if (!user) {
		fprintf(stderr, "Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;

	case 'F':
	    /* Sub-filter */
	    if (Settings.subFilter) {
		free((void *) Settings.subFilter);
	    }
	    Settings.subFilter = strdup(optarg);
	    if (!Settings.subFilter) {
		fprintf(stderr, "Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;

	case 'W':
	    /* Absolute minimum to wait between each slave's start-up */
	    if (sscanf(optarg, "%d", &Settings.waitTime) != 1) usage();
	    if (Settings.waitTime < 0) {
		Settings.waitTime = 0;
	    }
	    break;

	case 'w':
	    /* How long to wait between each slave's start-up */
	    if (sscanf(optarg, "%d", &Settings.slewTime) != 1) usage();
	    if (Settings.slewTime < 1) {
		Settings.slewTime = 1;
	    }
	    break;

	case 'p':
	    /* Write our pid to this file */
	    if (pidfile != NULL) free(pidfile);

	    pidfile = strdup(optarg);
	    if (!pidfile) {
		fprintf(stderr, "Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'f':
	    /* Filter program */
	    if (optarg[0] != '/') {
		fprintf(stderr, "-f: You must supply an absolute path for filter program\n");
		exit(EXIT_FAILURE);
	    }
	    Settings.progPath = strdup(optarg);
	    if (!Settings.progPath) {
		fprintf(stderr, "strdup: Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'u':
	    Settings.flushStats = 1;
	    break;
	case 't':
	    Settings.statsFile = strdup(optarg);
	    if (!Settings.statsFile) {
		fprintf(stderr, "strdup: Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	case 'T':
	    Settings.statsToSyslog = 1;
	    break;
	case 'l':
	    Settings.doSyslog = 1;
	    break;
	case 'd':
	    DebugEvents = 1;
	    break;
	case 'h':
	    usage();
	    break;
	case 'm':
	    if (sscanf(optarg, "%d", &Settings.minSlaves) != 1) usage();
	    if (Settings.minSlaves < 0) Settings.minSlaves = 0;
	    break;
	case 'x':
	    if (sscanf(optarg, "%d", &Settings.maxSlaves) != 1) usage();
	    break;
	case 'r':
	    if (sscanf(optarg, "%d", &Settings.maxRequests) != 1) usage();
	    if (Settings.maxRequests < 1) Settings.maxRequests = 1;
	    break;
	case 'i':
	    if (sscanf(optarg, "%d", &Settings.maxIdleTime) != 1) usage();
	    if (Settings.maxIdleTime < 10) Settings.maxIdleTime = 10;
	    break;
	case 'b':
	    if (sscanf(optarg, "%d", &Settings.busyTimeout) != 1) usage();
	    if (Settings.busyTimeout < 10) Settings.busyTimeout = 10;
	    break;
	case 'c':
	    if (sscanf(optarg, "%d", &Settings.clientTimeout) != 1) usage();
	    if (Settings.clientTimeout < 10) Settings.clientTimeout = 10;
	    break;
	case 's':
	    Settings.sockName = strdup(optarg);
	    if (!Settings.sockName) {
		fprintf(stderr, "strdup: Out of memory\n");
		exit(EXIT_FAILURE);
	    }
	    break;
	default:
	    fprintf(stderr, "\n");
	    usage();
	}
    }

    /* Drop privileges */
    if (user) {
	pw = getpwnam(user);
	if (!pw) {
	    fprintf(stderr, "Unknown user '%s'\n", user);
	    exit(EXIT_FAILURE);
	}
	if (drop_privs(user, pw->pw_uid, pw->pw_gid) < 0) {
	    exit(EXIT_FAILURE);
	}
	free(user);
    }

    /* Warn */
    if (!getuid() || !geteuid()) {
	fprintf(stderr,
		"ERROR: You must not run mimedefang-multiplexor as root.\n"
		"Use the -U option to set a non-root user.\n");
	exit(EXIT_FAILURE);
    }

    if (chdir(SPOOLDIR) < 0) {
	fprintf(stderr, "Unable to chdir(%s): %s\n",
		SPOOLDIR, strerror(errno));
	exit(EXIT_FAILURE);
    }

    /* Fix obvious stupidities */
    if (Settings.maxSlaves < 1) {
	Settings.maxSlaves = 1;
    }
    if (Settings.minSlaves > Settings.maxSlaves) {
	Settings.minSlaves = Settings.maxSlaves;
    }

    /* Allocate slaves and place them on the free list */
    AllSlaves = calloc(Settings.maxSlaves, sizeof(Slave));

    if (!AllSlaves) {
	fprintf(stderr, "Unable to allocate memory for slaves\n");
	exit(EXIT_FAILURE);
    }

    /* Make an event selector */
    es = Event_CreateSelector();
    if (!es) {
	fprintf(stderr, "Could not create event selector: %s\n",
		strerror(errno));
	exit(EXIT_FAILURE);
    }

    SlaveCount[STATE_STOPPED] = Settings.maxSlaves;
    SlaveCount[STATE_IDLE]    = 0;
    SlaveCount[STATE_BUSY]    = 0;
    SlaveCount[STATE_KILLED]  = 0;

    Slaves[STATE_STOPPED] = &AllSlaves[0];
    Slaves[STATE_IDLE]    = NULL;
    Slaves[STATE_BUSY]    = NULL;
    Slaves[STATE_KILLED]  = NULL;

    /* Set some fields in slaves */
    for (i=0; i<Settings.maxSlaves; i++) {
	Slave *s = &AllSlaves[i];
	s->es = es;
	s->pid = (pid_t) -1;
	s->slaveStdin = -1;
	s->slaveStdout = -1;
	s->slaveStderr = -1;
	s->clientFD = -1;
	s->oom = 0;
	s->event = NULL;
	s->errHandler = NULL;
	s->termHandler = NULL;
	s->workdir[0] = 0;
	s->generation = Generation;
	s->state = STATE_STOPPED;
    }

    /* Set up the linked list */
    for (i=0; i<Settings.maxSlaves-1; i++) {
	AllSlaves[i].next = &AllSlaves[i+1];
    }
    AllSlaves[Settings.maxSlaves-1].next = NULL;

    /* Make a UNIX-domain socket to talk to mimedefang */
    (void) remove(Settings.sockName);

    sock = socket(AF_LOCAL, SOCK_STREAM, 0);
    if (sock < 0) {
	fprintf(stderr, "Could not create UNIX-domain socket: %s\n",
		strerror(errno));
	exit(EXIT_FAILURE);
    }
    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, Settings.sockName, sizeof(addr.sun_path)-1);

    if (bind(sock, (struct sockaddr *) &addr, SUN_LEN(&addr)) < 0) {
	fprintf(stderr, "Could not bind UNIX-domain socket %s: %s\n",
		Settings.sockName, strerror(errno));
	exit(EXIT_FAILURE);
    }
    (void) chmod(Settings.sockName, 0600);
    if (listen(sock, 5) < 0) {
	fprintf(stderr, "Could not listen on UNIX-domain socket %s: %s\n",
		Settings.sockName, strerror(errno));
	exit(EXIT_FAILURE);
    }

    /* Set up an accept loop */
    if (!EventTcp_CreateAcceptor(es, sock, handleAccept)) {
	fprintf(stderr, "Could not make accept() handler: %s\n",
		strerror(errno));
	exit(EXIT_FAILURE);
    }

    /* Ignore sigpipe */
    signal(SIGPIPE, SIG_IGN);

    /* Set up pipe for signal handler for slave termination notification*/
    if (pipe(Pipe) < 0) {
	perror("pipe");
	exit(EXIT_FAILURE);
    }

    /* Create event handler for pipe */
    if (!Event_AddHandler(es, Pipe[0],
			  EVENT_FLAG_READABLE, handlePipe, NULL)) {
	fprintf(stderr, "Could not make pipe handler: %s\n",
		strerror(errno));
	exit(EXIT_FAILURE);
    }

    /* Daemonize */
    i = fork();
    if (i < 0) {
	fprintf(stderr, "fork() failed\n");
	exit(EXIT_FAILURE);
    } else if (i != 0) {
	/* parent */
	exit(EXIT_SUCCESS);
    }
    setsid();
    signal(SIGHUP, SIG_IGN);
    i = fork();
    if (i < 0) {
	fprintf(stderr, "fork() failed\n");
	exit(EXIT_FAILURE);
    } else if (i != 0) {
	exit(EXIT_SUCCESS);
    }

    /* Write pid */
    if (pidfile) {
	fp = fopen(pidfile, "w");
	if (fp) {
	    fprintf(fp, "%d\n", (int) getpid());
	    fclose(fp);
	}
	free(pidfile);
    }

    for (i=0; i<256; i++) {
	if (i == sock || i == Pipe[0] || i == Pipe[1]) continue;
	(void) close(i);
    }

    /* Direct stdin/stdout/stderr to /dev/null */
    open("/dev/null", O_RDWR);
    open("/dev/null", O_RDWR);
    open("/dev/null", O_RDWR);

    /* Syslog if required */
    openlog("mimedefang-multiplexor", 0, facility);

    /* Set signal handler for SIGCHLD */
    act.sa_handler = childHandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_NOCLDSTOP | SA_RESTART;
    if (sigaction(SIGCHLD, &act, NULL) < 0) {
	syslog(LOG_ERR, "sigaction failed - exiting: %m");
	exit(EXIT_FAILURE);
    }

    /* Set up SIGHUP handler */
    act.sa_handler = hupHandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_RESTART;
    if (sigaction(SIGHUP, &act, NULL) < 0) {
	syslog(LOG_ERR, "sigaction failed - exiting: %m");
	exit(EXIT_FAILURE);
    }

    /* Set up SIGINT handler */
    act.sa_handler = intHandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_RESTART;
    if (sigaction(SIGINT, &act, NULL) < 0) {
	syslog(LOG_ERR, "sigaction failed - exiting: %m");
	exit(EXIT_FAILURE);
    }

    if (DOLOG) {
	syslog(LOG_INFO, "started; minSlaves=%d, maxSlaves=%d, maxRequests=%d, maxIdleTime=%d, busyTimeout=%d, clientTimeout=%d",
	       Settings.minSlaves,
	       Settings.maxSlaves,
	       Settings.maxRequests,
	       Settings.maxIdleTime,
	       Settings.busyTimeout,
	       Settings.clientTimeout);
    }

    /* Open stats file */
    statsReopenFile();

    /* Kick off the starting of slaves */
    bringSlavesUpToMin(es, 0, 0, NULL);

    /* Set up a timer handler to check for idle timeouts */
    t.tv_usec = 0;
    t.tv_sec = Settings.maxIdleTime;
    Event_AddTimerHandler(es, t, handleIdleTimeout, NULL);

    /* Set up a timer handler to log status, if desired */
    if (Settings.logStatusInterval) {
	t.tv_usec = 0;
	t.tv_sec = Settings.logStatusInterval;
	Event_AddTimerHandler(es, t, doStatusLog, NULL);
    }
    if (DebugEvents) {
	Event_EnableDebugging("/var/log/mdefang-event-debug.log");
    }

    /* Set signal handler for SIGTERM */
    signal(SIGTERM, sigterm);

    /* And loop... */
    while(1) {
	if (Event_HandleEvent(es) < 0) {
	    syslog(LOG_CRIT, "Error in Event_HandleEvent: %m.  MULTIPLEXOR IS TERMINATING.");
	    sigterm(0);
	}
    }
}

/**********************************************************************
* %FUNCTION: handleSlaveStderr
* %ARGUMENTS:
*  es -- event selector
*  fd -- file descriptor ready for reading
*  flags -- flags from event-handling code
*  data -- pointer to slave
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Handles a readable stderr descriptor.  Reads data; if we're logging,
*  sends it to syslog; otherwise, discards data.
***********************************************************************/
static void
handleSlaveStderr(EventSelector *es,
		  int fd,
		  unsigned int flags,
		  void *data)
{
    Slave *s = (Slave *) data;
    char buffer[512];
    int n;

    while ( (n=read(fd, buffer, sizeof(buffer)-1)) > 0) {
	buffer[n] = 0;
	/* Heuristic... Perl spits this out, I think*/
	if (strstr(buffer, "Out of memory!")) {
	    s->oom = 1;
	}

	if (DOLOG) {
	    syslog(LOG_INFO, "Slave %d stderr: %s", SLAVENO(s), buffer);
	}
    }
    if (n == 0 || (n < 0 && errno != EAGAIN)) {
	/* EOF or error reading stderr -- close it and cancel handler */
	if (n < 0) {
	    if (DOLOG) {
		syslog(LOG_WARNING,
		       "handleSlaveStderr: Error reading from slave %d's stderr: %m", SLAVENO(s));
	    }
	}
	close(s->slaveStderr);
	Event_DelHandler(s->es, s->errHandler);
	s->errHandler = NULL;
	s->slaveStderr = -1;
    }
}

/**********************************************************************
* %FUNCTION: handleAccept
* %ARGUMENTS:
*  es -- event selector
*  fd -- accepted connection
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Handles a connection attempt from MIMEDefang.  Sets up a command-reader.
***********************************************************************/
static void
handleAccept(EventSelector *es, int fd)
{
    if (!EventTcp_ReadBuf(es, fd, MAX_CMD_LEN, '\n', handleCommand,
			  Settings.clientTimeout, 1, NULL)) {
	if (DOLOG) {
	    syslog(LOG_ERR, "handleAccept: EventTcp_ReadBuf failed: %m");
	}
	close(fd);
    }
}

/**********************************************************************
* %FUNCTION: handleCommand
* %ARGUMENTS:
*  es -- event selector
*  fd -- connection to MIMEDefang
*  buf -- buffer of data read from MIMEDefang
*  len -- amount of data read from MIMEDefang
*  flag -- flag from reader
*  data -- ignored
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Handles a MIMEDefang command.  Currently, only two commands are
*  available:
*  "free\n" --> Returns number of free slaves.
*  "scan dir\n" --> Scans directory, returns "done" when done or "busy" if
*      no free slaves or "error: msg" if an error occurred.
***********************************************************************/
static void
handleCommand(EventSelector *es,
	      int fd,
	      char *buf,
	      int len,
	      int flag,
	      void *data)
{
    char answer[MAX_CMD_LEN];

    if (flag == EVENT_TCP_FLAG_TIMEOUT || flag == EVENT_TCP_FLAG_IOERROR) {
	if (DOLOG) {
	    syslog(LOG_DEBUG, "handleCommand: Timeout or error: Flag = %d",
		   flag);
	}
	/* Client timeout or error */
	close(fd);
	return;
    }

    /* Null-terminate buffer */
    if (len) {
	buf[len-1] = 0;
	len--;
    }

    if (len == 4 && !strcmp(buf, "free")) {
	snprintf(answer, sizeof(answer), "%d\n",
		 NUM_FREE_SLAVES);
	reply_to_mimedefang(es, fd, answer);
	return;
    }

    if (len == 6 && !strcmp(buf, "status")) {
	doStatus(es, fd);
	return;
    }

    if (len == 4 && !strcmp(buf, "msgs")) {
	snprintf(answer, sizeof(answer), "%d\n", NumMsgsProcessed);
	reply_to_mimedefang(es, fd, answer);
	return;
    }

    if (len == 6 && !strcmp(buf, "reread")) {
	newGeneration();
	reply_to_mimedefang(es, fd, "Forced reread of filter rules\n");
	return;
    }

    if (len > 5 && !strncmp(buf, "scan ", 5)) {
	doScan(es, fd, buf);
	return;
    }

    if (len > 8 && !strncmp(buf, "relayok ", 8)) {
	doSlaveCommand(es, fd, buf);
	return;
    }
    if (len > 9 && !strncmp(buf, "senderok ", 9)) {
	doSlaveCommand(es, fd, buf);
	return;
    }
    if (len > 8 && !strncmp(buf, "recipok ", 8)) {
	doSlaveCommand(es, fd, buf);
	return;
    }

    /* Unknown command */
    if (DOLOG) {
	syslog(LOG_ERR, "Unknown command: %s", buf);
    }
    snprintf(answer, sizeof(answer), "error: Unknown command %s\n",
	     buf);
    reply_to_mimedefang(es, fd, answer);
}

/**********************************************************************
* %FUNCTION: doScan
* %ARGUMENTS:
*  es -- event selector
*  fd -- connection to MIMEDefang
*  cmd -- scan command
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Sends the scan command to the slave and arranges for answer to be
*  sent back when scanning is complete.
***********************************************************************/
static void
doScan(EventSelector *es, int fd, char *cmd)
{
    Slave *s;
    int len;

    /* Add newline back to command */
    len = strlen(cmd);
    if (len < MAX_CMD_LEN-1) {
	cmd[len+1] = 0;
	cmd[len] = '\n';
    } else {
	char *answer = "error: Command too long\n";
	reply_to_mimedefang(es, fd, answer);
	if (DOLOG) {
	    syslog(LOG_DEBUG, "doScan: Command too long");
	}
	return;
    }

    /* Find a free slave */
    s = findFreeSlave();
    if (!s) {
	char *answer = "error: No free slaves\n";
	if (DOLOG) {
	    syslog(LOG_WARNING, "No free slaves");
	}
	reply_to_mimedefang(es, fd, answer);
	return;
    }

    if (activateSlave(s, "About to perform scan") == (pid_t) -1) {
	char *answer = "error: Unable to activate slave\n";
	if (DOLOG) {
	    syslog(LOG_ERR, "Unable to activate slave %d",
		   SLAVENO(s));
	}
	reply_to_mimedefang(es, fd, answer);
	return;
    }

    /* Put the slave on the busy list */
    putOnList(s, STATE_BUSY);

    /* Set slave's clientFD so we can reply */
    s->clientFD = fd;

    /* Set slave's working directory, sans newline */
    snprintf(s->workdir, sizeof(s->workdir), "%s", &cmd[strlen("scan ")]);
    s->workdir[MAX_DIR_LEN-1] = 0;
    len = strlen(s->workdir);
    if (len && s->workdir[len-1] == '\n') {
	s->workdir[len-1] = 0;
    }

    /* And tell the slave to go ahead... */
    s->event = EventTcp_WriteBuf(es, s->slaveStdin, cmd, strlen(cmd),
				 handleSlaveReceivedCommand,
				 Settings.clientTimeout, s);
    if (!s->event) {
	if (DOLOG) syslog(LOG_ERR, "doScan: EventTcp_WriteBuf failed: %m");
	killSlave(s, "EventTcp_WriteBuf failed");
    } else {
	statsLog("StartFilter", SLAVENO(s), NULL);
    }
}

/**********************************************************************
* %FUNCTION: doSlaveCommand
* %ARGUMENTS:
*  es -- event selector
*  fd -- connection to MIMEDefang
*  cmd -- command
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Sends the command to the slave and arranges for answer to be
*  sent back when command finishes.
***********************************************************************/
static void
doSlaveCommand(EventSelector *es, int fd, char *cmd)
{
    Slave *s;
    int len;
    char reason[200];

    sprintf(reason, "About to execute command '%.100s'", cmd);

    /* Add newline back to command */
    len = strlen(cmd);
    if (len < MAX_CMD_LEN-1) {
	cmd[len+1] = 0;
	cmd[len] = '\n';
    } else {
	char *answer = "error: Command too long\n";
	reply_to_mimedefang(es, fd, answer);
	if (DOLOG) {
	    syslog(LOG_DEBUG, "doSlaveCommand: Command too long");
	}
	return;
    }

    /* Find a free slave */
    s = findFreeSlave();
    if (!s) {
	char *answer = "error: No free slaves\n";
	if (DOLOG) {
	    syslog(LOG_WARNING, "No free slaves");
	}
	reply_to_mimedefang(es, fd, answer);
	return;
    }

    if (activateSlave(s, reason) == (pid_t) -1) {
	char *answer = "error: Unable to activate slave\n";
	if (DOLOG) {
	    syslog(LOG_ERR, "Unable to activate slave %d",
		   SLAVENO(s));
	}
	reply_to_mimedefang(es, fd, answer);
	return;
    }

    /* Put the slave on the busy list */
    putOnList(s, STATE_BUSY);

    /* Set slave's clientFD so we can reply */
    s->clientFD = fd;

    /* Null workdir signals not to log EndFilter event */
    s->workdir[0] = 0;

    /* And tell the slave to go ahead... */
    s->event = EventTcp_WriteBuf(es, s->slaveStdin, cmd, strlen(cmd),
				 handleSlaveReceivedCommand,
				 Settings.clientTimeout, s);
    if (!s->event) {
	if (DOLOG) syslog(LOG_ERR, "doSlaveCommand: EventTcp_WriteBuf failed: %m");
	killSlave(s, "EventTcp_WriteBuf failed");
    }
}

/**********************************************************************
* %FUNCTION: handleSlaveReceivedCommand
* %ARGUMENTS:
*  es -- event selector
*  fd -- connection to MIMEDefang
*  buf -- buffer of data read from MIMEDefang
*  len -- amount of data read from MIMEDefang
*  flag -- flag from reader
*  data -- the slave
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Called when command has been written to slave.  Sets up an event handler
*  to read the slave's reply
***********************************************************************/
static void
handleSlaveReceivedCommand(EventSelector *es,
			   int fd,
			   char *buf,
			   int len,
			   int flag,
			   void *data)
{
    Slave *s = (Slave *) data;

    /* Event was triggered */
    s->event = NULL;

    if (flag == EVENT_TCP_FLAG_TIMEOUT || flag == EVENT_TCP_FLAG_IOERROR) {
	/* Error writing to slave */
	char *answer = "error: Error talking to slave process\n";
	if (DOLOG) {
	    syslog(LOG_DEBUG, "handleSlaveReceivedCommand(%d): Timeout or error: Flag = %d",
		   SLAVENO(s),
		   flag);
	}
	reply_to_mimedefang(es, s->clientFD, answer);
	/* The reply_to_mimedefang will close clientFD when it's done */
	s->clientFD = -1;

	/* Kill the slave process */
	killSlave(s, "Error talking to slave process");
	return;
    }

    /* Slave has been given the command; now wait for it to reply */
    s->event = EventTcp_ReadBuf(es, s->slaveStdout, MAX_CMD_LEN, '\n',
				handleSlaveReceivedAnswer,
				Settings.busyTimeout, 1, s);
    if (!s->event) {
	if (DOLOG) syslog(LOG_ERR, "handleSlaveReceivedCommand: EventTcp_ReadBuf failed: %m");
	killSlave(s, "EventTcp_ReadBuf failed");
    }
}

/**********************************************************************
* %FUNCTION: handleSlaveReceivedAnswer
* %ARGUMENTS:
*  es -- event selector
*  fd -- connection to MIMEDefang
*  buf -- buffer of data read from MIMEDefang
*  len -- amount of data read from MIMEDefang
*  flag -- flag from reader
*  data -- the slave
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Called when the slave's answer comes back.
***********************************************************************/
static void
handleSlaveReceivedAnswer(EventSelector *es,
			  int fd,
			  char *buf,
			  int len,
			  int flag,
			  void *data)
{
    Slave *s = (Slave *) data;

    /* Event was triggered */
    s->event = NULL;

    /* If nothing was received from slave, send error message back */
    if (!len || (flag == EVENT_TCP_FLAG_TIMEOUT)) {
	if (flag == EVENT_TCP_FLAG_TIMEOUT) {
	    /* Heuristic... */
	    if (SlaveCount[STATE_BUSY] > 3) {
		reply_to_mimedefang(es, s->clientFD,
				    "ERR Filter timed out - system may be overloaded "
				    "(consider increasing busy timeout)\n");
		/* The reply_to_mimedefang will close clientFD when it's done */
		s->clientFD = -1;
	    } else {
		reply_to_mimedefang(es, s->clientFD,
				    "ERR Filter timed out - check filter rules or system load\n");
		/* The reply_to_mimedefang will close clientFD when it's done */
		s->clientFD = -1;
	    }
	} else {
	    if (DOLOG) {
		if (s->oom) {
		    syslog(LOG_ERR, "Slave %d ran out of memory -- possible DoS attack due to complex MIME?", SLAVENO(s));
		} else {
		    syslog(LOG_ERR, "Slave %d died prematurely -- check your filter rules", SLAVENO(s));
		}
	    } else {
		if (s->oom) {
		    syslog(LOG_ERR, "Slave %d ran out of memory -- possible DoS attack due to complex MIME?", SLAVENO(s));
		} else {
		    syslog(LOG_ERR, "Slave %d died prematurely -- check your filter rules and use the '-l' flag on mimedefang-multiplexor to see Perl error messages", SLAVENO(s));
		}
	    }
	    reply_to_mimedefang(es, s->clientFD, "ERR No response from slave\n");
	    /* The reply_to_mimedefang will close clientFD when it's done */
	    s->clientFD = -1;
	}
    } else {
	/* Write the slave's answer back to the client */
	reply_to_mimedefang_with_len(es, s->clientFD, buf, len);
	/* The reply_to_mimedefang will close clientFD when it's done */
	s->clientFD = -1;
    }

    s->numRequests++;

    if (s->workdir[0]) {
	/* Only increment NumMsgsProcessed for a "scan" command */
	NumMsgsProcessed++;
	s->numScans++;
    }

    /* If we had a busy timeout, kill the slave */
    if (flag == EVENT_TCP_FLAG_TIMEOUT) {
	killSlave(s, "Busy timeout");
    } else {
	/* Put slave on free list */
	putOnList(s, STATE_IDLE);

	/* Record time when this slave became idle */
	s->idleTime = time(NULL);

	/* Check slave for expiry */
	checkSlaveForExpiry(s);
    }
    if (s->workdir[0]) {
	statsLog("EndFilter", SLAVENO(s), "numRequests=%d", s->numRequests);
    }
}

/**********************************************************************
* %FUNCTION: unlinkFromList
* %ARGUMENTS:
*  s -- a slave
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Unlinks "s" from the list it is currently on.
***********************************************************************/
static void
unlinkFromList(Slave *s)
{
    int state = s->state;
    Slave *prev = NULL;

    if (!Slaves[state]) {
	/* Panic... slave not on this list */
	syslog(LOG_CRIT, "%s slave %d not found on its list!",
	       state_name(s), SLAVENO(s));
	return;
    }

    /* If it's the first on the list, simple */
    if (Slaves[state] == s) {
	Slaves[state] = s->next;
	s->next = NULL;
	return;
    }

    /* Somewhere in the middle of the list */
    prev = Slaves[state];
    while (prev->next != s) {
	if (!prev->next) {
	    /* Panic... slave not on this list */
	    syslog(LOG_CRIT, "%s slave %d not found on its list!",
		   state_name(s), SLAVENO(s));
	    return;
	}
	prev = prev->next;
    }

    prev->next = s->next;
    s->next = NULL;
}

/**********************************************************************
* %FUNCTION: putOnList
* %ARGUMENTS:
*  s -- a slave
*  state -- state to change to.
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Sets s->state to state, and moves s to the Slaves[state] list.
***********************************************************************/
static void
putOnList(Slave *s, int state)
{
    if (s->state == state) {
	/* Already there, nothing to do */
	return;
    }

    /* Adjust counts */
    SlaveCount[s->state]--;
    SlaveCount[state]++;

    unlinkFromList(s);
    s->next = Slaves[state];
    Slaves[state] = s;
    s->state = state;
}

/**********************************************************************
* %FUNCTION: activateSlave
* %ARGUMENTS:
*  s -- a slave
*  reason -- reason slave is being activated
* %RETURNS:
*  The process-ID of the slave
* %DESCRIPTION:
*  Activates the slave if it is not currently associated with a running
*  process.
***********************************************************************/
static pid_t
activateSlave(Slave *s, char const *reason)
{
    int pin[2], pout[2], perr[2];
    char const *pname;
    int i;
    int flags;
    time_t now = (time_t) 0; /* Avoid compiler warning by initializing */
    sigset_t sigs;

    /* Check if it's already active */
    if (s->state == STATE_BUSY ||
	s->state == STATE_IDLE) {
	return s->pid;
    }

    /* Check if enough time has elapsed */
    if (Settings.waitTime) {
	now = time(NULL);
	if (LastSlaveActivation &&
	     (((int) (now - LastSlaveActivation)) < Settings.waitTime)) {
	    if (DOLOG) {
		syslog(LOG_DEBUG, "Did not start slave %d: Not enough time elapsed since last slave activation", SLAVENO(s));
	    }
	    return -1;
	}
    }

    /* Set up pipes */
    if (pipe(pin) < 0) {
	if (DOLOG) syslog(LOG_ERR, "Could not start slave %d: pipe failed: %m",
			  SLAVENO(s));
	return -1;
    }
    if (pipe(pout) < 0) {
	if (DOLOG) syslog(LOG_ERR, "Could not start slave %d: pipe failed: %m",
			  SLAVENO(s));
	close(pin[0]);
	close(pin[1]);
	return -1;
    }

    if (pipe(perr) < 0) {
	if (DOLOG) syslog(LOG_ERR, "Could not start slave %d: pipe failed: %m",
			  SLAVENO(s));
	close(pin[0]);
	close(pin[1]);
	close(pout[0]);
	close(pout[1]);
	return -1;
    }

    /* fork and exec */
    s->pid = fork();

    if (s->pid == (pid_t) -1) {
	if (DOLOG) syslog(LOG_ERR, "Could not start slave %d: fork failed: %m",
			  SLAVENO(s));
	close(pin[0]);
	close(pin[1]);
	close(pout[0]);
	close(pout[1]);
	close(perr[0]);
	close(perr[1]);
	return s->pid; /* Fork failed */
    }

    if (s->pid) {
	putOnList(s, STATE_IDLE);
	/* Record time when this slave became idle */
	s->idleTime = time(NULL);

	/* In the parent -- return */
	close(pin[0]);
	close(pout[1]);
	close(perr[1]);
	s->slaveStdin = pin[1];
	s->slaveStdout = pout[0];
	s->slaveStderr = perr[0];
	s->activated = Activations++;

	/* Make slaveStderr non-blocking */
	flags = fcntl(s->slaveStderr, F_GETFL, 0);
	if (flags != -1) {
	    flags = fcntl(s->slaveStderr, F_SETFL, flags | O_NONBLOCK);
	}

	if (flags == -1) {
	    /* Doh.. next to impossible to recover from this error */
	    syslog(LOG_ERR, "Could not make slave %d's stderr non-blocking: %m", SLAVENO(s));
	}

	/* Handle anything written to slave's stderr */
	s->errHandler = Event_AddHandler(s->es, s->slaveStderr,
					 EVENT_FLAG_READABLE,
					 handleSlaveStderr, s);
	s->clientFD = -1;
	s->numRequests = 0;
	s->numScans = 0;
	s->oom = 0;
	s->generation = Generation;
	if (DOLOG) {
	    syslog(LOG_INFO, "Starting slave %d (pid %lu) (%d running): %s",
		   SLAVENO(s),
		   (unsigned long) s->pid, NUM_RUNNING_SLAVES, reason);
	}
	statsLog("StartSlave", SLAVENO(s), "reason=\"%s\"", reason);
	if (Settings.waitTime) {
	    LastSlaveActivation = now;
	}
	return s->pid;
    }

    /* In the child */

    /* Set resource limits */
#ifdef HAVE_SETRLIMIT
    limit_mem_usage(Settings.maxRSS, Settings.maxAS);
#endif

    close(pin[1]);
    close(pout[0]);
    close(perr[0]);
    dup2(pin[0], STDIN_FILENO);
    dup2(pout[1], STDOUT_FILENO);
    dup2(perr[1], STDERR_FILENO);

    if (pin[0] != STDIN_FILENO) close(pin[0]);
    if (pout[1] != STDOUT_FILENO) close(pout[1]);
    if (perr[1] != STDERR_FILENO) close(perr[1]);

    /* Close unneeded file descriptors */
    for (i=STDERR_FILENO+1; i<256; i++) {
	(void) close(i);
    }

    pname = strrchr(Settings.progPath, '/');
    if (pname) {
	pname++;
    } else {
	pname = Settings.progPath;
    }
    signal(SIGTERM, SIG_DFL);
    signal(SIGCHLD, SIG_DFL);
    signal(SIGHUP, SIG_DFL);
    signal(SIGINT, SIG_DFL);
    sigemptyset(&sigs);
    sigprocmask(SIG_BLOCK, &sigs, NULL);

    if (Settings.subFilter) {
	execl(Settings.progPath, pname, "-f", Settings.subFilter, "-server", NULL);
    } else {
	execl(Settings.progPath, pname, "-server", NULL);
    }
    _exit(EXIT_FAILURE);
}

/**********************************************************************
* %FUNCTION: checkSlaveForExpiry
* %ARGUMENTS:
*  s -- a slave to check
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  If the slave has served too many requests, it is killed.
***********************************************************************/
static void
checkSlaveForExpiry(Slave *s)
{
    if (s->numRequests >= Settings.maxRequests) {
	char reason[200];
	snprintf(reason, sizeof(reason), "Slave has processed %d requests",
		 Settings.maxRequests);
	killSlave(s, reason);
    } else if (s->generation < Generation) {
	killSlave(s, "New generation -- forcing reread of filter rules");
    }
}

/**********************************************************************
* %FUNCTION: killSlave
* %ARGUMENTS:
*  s -- a slave to kill
*  reason -- reason slave is being killed
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Kills a slave by closing its pipe.  The slave then has 10 seconds
*  to clean up and exit before we send SIGTERM
***********************************************************************/
static void
killSlave(Slave *s, char const *reason)
{
    struct timeval t;
    int withPrejudice = 0;

    if (s->state != STATE_KILLED) {
	if (DOLOG) syslog(LOG_INFO, "Killing %s slave %d (pid %lu): %s",
			  state_name_lc(s),
			  SLAVENO(s),
			  (unsigned long) s->pid, reason);
	/* In case, for some weird reason, the slave has stopped... */
	kill(s->pid, SIGCONT);

	/* If slave is busy, we kill it with prejudice */
	if (s->state == STATE_BUSY) {
	    withPrejudice = 1;
	    kill(s->pid, SIGTERM);
	}

	putOnList(s, STATE_KILLED);
	/* Close stdin so slave sees EOF */
	close(s->slaveStdin);
	s->slaveStdin = -1;

	/* Kill any pending event */
	if (s->event) {
	    EventTcp_CancelPending(s->event);
	    s->event = NULL;
	}

	/* Set up a timer to send SIGTERM if slave doesn't exit in
	   10 seconds */
	t.tv_sec = 10;
	t.tv_usec = 0;
	if (withPrejudice) {
	    s->termHandler = Event_AddTimerHandler(s->es, t,
						   nukeSlave, (void *) s);
	} else {
	    s->termHandler = Event_AddTimerHandler(s->es, t,
						   terminateSlave, (void *) s);
	}
	statsLog("KillSlave", SLAVENO(s), "reason=\"%s\"", reason);
    }
}

/**********************************************************************
* %FUNCTION: terminateSlave
* %ARGUMENTS:
*  es -- Event selector
*  fd -- not used
*  flags -- ignored
*  data -- the slave to terminate
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Sends a SIGTERM to slave, because it's taken too long to exit.
***********************************************************************/
static void
terminateSlave(EventSelector *es,
	       int fd,
	       unsigned int flags,
	       void *data)
{
    struct timeval t;
    Slave *s = (Slave *) data;
    s->termHandler = NULL;
    if (s->pid != (pid_t) -1) {
	if (DOLOG) {
	    syslog(LOG_INFO,
		   "Slave %d (pid %lu) taking too long to exit; sending SIGTERM",
		   SLAVENO(s), (unsigned long) s->pid);
	}
	kill(s->pid, SIGCONT);
	kill(s->pid, SIGTERM);
	/* Set up a timer to send SIGKILL if slave doesn't exit in
	   10 seconds */
	t.tv_sec = 10;
	t.tv_usec = 0;
	s->termHandler = Event_AddTimerHandler(s->es, t, nukeSlave, (void *) s);
    }
}

/**********************************************************************
* %FUNCTION: nukeSlave
* %ARGUMENTS:
*  es -- Event selector
*  fd -- not used
*  flags -- ignored
*  data -- the slave to nuke
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Sends a SIGKILL to slave, because it's taken too long to exit, in spite
*  of getting a SIGTERM
***********************************************************************/
static void
nukeSlave(EventSelector *es,
	  int fd,
	  unsigned int flags,
	  void *data)
{
    Slave *s = (Slave *) data;
    s->termHandler = NULL;
    if (s->pid != (pid_t) -1) {
	if (DOLOG) {
	    syslog(LOG_INFO,
		   "Slave %d (pid %lu) taking way too long to exit; sending SIGKILL",
		   SLAVENO(s), (unsigned long) s->pid);
	}
	kill(s->pid, SIGCONT);
	kill(s->pid, SIGKILL);
    }
}

/**********************************************************************
*%FUNCTION: childHandler
*%ARGUMENTS:
* sig -- signal number
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Called by SIGCHLD.  Writes 'C' to Pipe to wake up the select
* loop and cause reaping of dead sessions
***********************************************************************/
static void
childHandler(int sig)
{
    char byte = 'C';
    if (!ReapPending) {
	ReapPending = 1;
	if (!CharPending) {
	    int errno_save = errno;
	    write(Pipe[1], &byte, 1);
	    errno = errno_save;
	    CharPending = 1;;
	}
    }
}

/**********************************************************************
*%FUNCTION: hupHandler
*%ARGUMENTS:
* sig -- signal number
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Called by SIGHUP.  Writes 'H' to Pipe to wake up the select
* loop and cause closing and reopening of stats descriptor
***********************************************************************/
static void
hupHandler(int sig)
{
    char byte = 'H';
    if (!HupPending) {
	HupPending = 1;
	if (!CharPending) {
	    int errno_save = errno;
	    write(Pipe[1], &byte, 1);
	    errno = errno_save;
	    CharPending = 1;;
	}
    }
}

/**********************************************************************
*%FUNCTION: intHandler
*%ARGUMENTS:
* sig -- signal number
*%RETURNS:
* Nothing
*%DESCRIPTION:
* Called by SIGINT.  Writes 'I' to Pipe to wake up the select
* loop and cause closing and reopening of stats descriptor
***********************************************************************/
static void
intHandler(int sig)
{
    char byte = 'I';
    if (!IntPending) {
	IntPending = 1;
	if (!CharPending) {
	    int errno_save = errno;
	    write(Pipe[1], &byte, 1);
	    errno = errno_save;
	    CharPending = 1;;
	}
    }
}

/**********************************************************************
* %FUNCTION: handlePipe
* %ARGUMENTS:
*  es -- event selector (ignored)
*  fd -- file descriptor which is readable
*  flags -- ignored
*  data -- ignored
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Simply reads from the pipe; then reaps and/or closes and reopens
*  stats file.
***********************************************************************/
static void
handlePipe(EventSelector *es,
	    int fd,
	    unsigned int flags,
	    void *data)
{
    char buf[64];
    sigset_t sigs, osigs;
    int n;
    int doreap, dohup, doint;

    /* The read and the reset of CharPending must be atomic with respect
       to signal handlers. */
    sigemptyset(&sigs);
    sigaddset(&sigs, SIGCHLD);
    sigaddset(&sigs, SIGINT);
    sigaddset(&sigs, SIGHUP);

    n = sigprocmask(SIG_BLOCK, &sigs, &osigs);
    if (n < 0) {
	syslog(LOG_ERR, "sigprocmask failed: %m");
    }

    /* Begin atomic portion */
    read(fd, buf, 64);
    doreap = ReapPending;
    doint  = IntPending;
    dohup  = HupPending;
    CharPending = 0;
    ReapPending = 0;
    IntPending  = 0;
    HupPending  = 0;
    /* End atomic portion */

    /* Reset signal mask */
    if (n >= 0) {
	sigprocmask(SIG_SETMASK, &osigs, NULL);
    }

    if (doreap) {
	reapTerminatedSlaves();

	/* Activate new slaves if we've fallen below minimum */
	if (NUM_RUNNING_SLAVES < Settings.minSlaves) {
	    scheduleBringSlavesUpToMin(es);
	}
    }
    if (dohup) {
	statsReopenFile();
    }
    if (doint) {
	newGeneration();
    }
}

/**********************************************************************
* %FUNCTION: reapTerminatedSlaves
* %ARGUMENTS:
*  None
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Reaps all terminated slaves
***********************************************************************/
static void
reapTerminatedSlaves(void)
{
    pid_t pid;
    int status;
    Slave *s;

#ifdef HAVE_WAIT3
    struct rusage resource;

    while ((pid = wait3(&status, WNOHANG, &resource)) > 0) {
#else
    while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
#endif
	s = findSlaveByPid(pid);
	if (!s) continue;

	if (DOLOG) logSlaveReaped(s, status);
#ifdef HAVE_WAIT3
	log_slave_resource_usage(s, &resource);
#endif
	s->pid = (pid_t) -1;
	shutDescriptors(s);
	putOnList(s, STATE_STOPPED);
	statsLog("ReapSlave", SLAVENO(s), NULL);
    }
}

/**********************************************************************
* %FUNCTION: shutDescriptors
* %ARGUMENTS:
*  s -- a slave
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Closes slave's descriptors if necessary
***********************************************************************/
static void
shutDescriptors(Slave *s)
{
    if (s->slaveStdin >= 0) {
	close(s->slaveStdin);
	s->slaveStdin = -1;
    }
    if (s->slaveStdout >= 0) {
	close(s->slaveStdout);
	s->slaveStdout = -1;
    }
    if (s->slaveStderr >= 0) {
	close(s->slaveStderr);
	Event_DelHandler(s->es, s->errHandler);
	s->errHandler = NULL;
	s->slaveStderr = -1;
    }
    if (s->termHandler) {
	Event_DelHandler(s->es, s->termHandler);
	s->termHandler = NULL;
    }

    if (s->event) {
	EventTcp_CancelPending(s->event);
	s->event = NULL;
    }

    if (s->clientFD >= 0) {
	close(s->clientFD);
	s->clientFD = -1;
    }
}

/**********************************************************************
* %FUNCTION: handleIdleTimeout
* %ARGUMENTS:
*  es -- event selector
*  fd, flags, data -- ignored
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Called periodically to check for idle slaves that can be killed
***********************************************************************/
static void
handleIdleTimeout(EventSelector *es,
		  int fd,
		  unsigned int flags,
		  void *data)
{
    time_t now = time(NULL);
    struct timeval t;

    /* Kill idle slaves which have been idle for maxIdleTime */
    Slave *s = Slaves[STATE_IDLE];
    Slave *next;
    int numAlive = NUM_RUNNING_SLAVES;
    while(s && numAlive > Settings.minSlaves) {
	next = s->next;
	if (((unsigned long) now - (unsigned long) s->idleTime)
	    >= Settings.maxIdleTime) {
	    numAlive--;
	    killSlave(s, "Idle timeout");
	}
	s = next;
    }

    /* If there are fewer running slaves than Settings.minSlaves,
       then start some more.  Should never happen, but... */
    if (numAlive < Settings.minSlaves) {
	scheduleBringSlavesUpToMin(es);
    }

    /* Reschedule timer */
    t.tv_usec = 0;
    t.tv_sec = Settings.maxIdleTime;
    Event_AddTimerHandler(es, t, handleIdleTimeout, NULL);
}

/**********************************************************************
* %FUNCTION: doStatusLog
* %ARGUMENTS:
*  es -- event selector
*  fd, flags, data -- ignored
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Called periodically to log slave status (numRunning, etc.)
***********************************************************************/
static void
doStatusLog(EventSelector *es,
	    int fd,
	    unsigned int flags,
	    void *data)
{
    struct timeval t;

    char buf[1024];

    snprintf(buf, sizeof(buf),
	     "Slave status: Stopped=%d Idle=%d Busy=%d Killed=%d Msgs=%d Activations=%u",
	     SlaveCount[STATE_STOPPED],
	     SlaveCount[STATE_IDLE],
	     SlaveCount[STATE_BUSY],
	     SlaveCount[STATE_KILLED],
	     NumMsgsProcessed,
	     Activations);
    syslog(LOG_INFO, buf);

    /* Reschedule timer */
    t.tv_usec = 0;
    t.tv_sec = Settings.logStatusInterval;
    Event_AddTimerHandler(es, t, doStatusLog, NULL);
}

/**********************************************************************
* %FUNCTION: logSlaveReaped
* %ARGUMENTS:
*  s -- slave
*  status -- termination status
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Logs to syslog
***********************************************************************/
static void
logSlaveReaped(Slave *s, int status)
{
    int level;
    char *extra;
    char const *state;

    state = state_name(s);

    if (s->state == STATE_KILLED) {
	level = LOG_DEBUG;
	extra = "";
    } else {
	level = LOG_ERR;
	extra = " (SLAVE DIED UNEXPECTEDLY)";
    }
    if (WIFEXITED(status)) {
	syslog(level, "Reap: %s slave %d (pid %lu) exited normally with status %d%s",
	       state,
	       SLAVENO(s), (unsigned long) s->pid, WEXITSTATUS(status),
	       extra);
	return;
    }
    if (WIFSIGNALED(status)) {
	if (s->state == STATE_KILLED && (WTERMSIG(status) == SIGTERM ||
			  WTERMSIG(status) == SIGKILL)) {
	    syslog(level, "Reap: %s slave %d (pid %lu) exited due to SIGTERM/SIGKILL as expected.",
		   state,
		   SLAVENO(s), (unsigned long) s->pid);
	} else {
	    syslog(level, "Reap: %s slave %d (pid %lu) exited due to signal %d%s",
		   state,
		   SLAVENO(s), (unsigned long) s->pid, WTERMSIG(status),
		   extra);
	}
	return;
    }
    syslog(level, "Reap: %s slave %d (pid %lu) exited for unknown reason%s",
	   state,
	   SLAVENO(s), (unsigned long) s->pid, extra);
}

/**********************************************************************
* %FUNCTION: sigterm
* %ARGUMENTS:
*  sig -- signal number
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Called when SIGTERM received -- kills slaves and exits
***********************************************************************/
static void
sigterm(int sig)
{
    int i;

    if (DOLOG) {
	if (sig) {
	    syslog(LOG_INFO, "Received SIGTERM: Killing slaves and terminating");
	}
    }
    for (i=0; i<Settings.maxSlaves; i++) {
	if (AllSlaves[i].pid != (pid_t) -1) {
	    kill(AllSlaves[i].pid, SIGCONT);
	    kill(AllSlaves[i].pid, SIGTERM);
	}
    }
    exit(1);
}

/**********************************************************************
* %FUNCTION: statsReopenFile
* %ARGUMENTS:
*  None
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Opens or re-opens statistics file.
***********************************************************************/
static void
statsReopenFile(void)
{
    if (!Settings.statsFile) return;
    if (Settings.statsFP) {
	if (fclose(Settings.statsFP) == EOF) {
	    syslog(LOG_ERR, "Failed to close stats file: %m");
	}
	Settings.statsFP = NULL;
    }

    Settings.statsFP = fopen(Settings.statsFile, "a");
    if (!Settings.statsFP) {
	syslog(LOG_ERR, "Could not open stats file %s: %m",
	       Settings.statsFile);
    }
}

/**********************************************************************
* %FUNCTION: statsLog
* %ARGUMENTS:
*  event -- name of event to put in log
*  slaveno -- slave involved in this event, or -1 if not slave-specific
*  fmt   -- "printf" format string with extra args to add to line
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Logs an event to the stats file
***********************************************************************/
static void
statsLog(char const *event, int slaveno, char const *fmt, ...)
{
    struct timeval now;
    time_t tnow;
    struct tm *t;
    char tbuf[64];
    char statbuf[1024];
    va_list ap;
    int ms;

    if (!Settings.statsFP && !Settings.statsToSyslog) return;

    gettimeofday(&now, NULL);
    tnow = (time_t) now.tv_sec;
    t = localtime(&tnow);
    ms = now.tv_usec / 1000;

    strftime(tbuf, sizeof(tbuf), "%d/%m/%Y:%H:%M:%S", t);
    snprintf(statbuf, sizeof(statbuf),
	     "%s %lu.%03d %s slave=%d nslaves=%d nbusy=%d",
	     tbuf, (unsigned long) tnow, ms, event, slaveno,
	     NUM_RUNNING_SLAVES, SlaveCount[STATE_BUSY]);

    statbuf[sizeof(statbuf)-1] = 0;

    if (fmt) {
	int len = strlen(statbuf);
	statbuf[len] = ' ';
	va_start(ap, fmt);
	vsnprintf(statbuf+len+1, sizeof(statbuf)-len-1, fmt, ap);
	va_end(ap);
	statbuf[sizeof(statbuf)-1] = 0;
    }

    if (Settings.statsFP) {
	fprintf(Settings.statsFP, "%s\n", statbuf);
	if (Settings.flushStats) fflush(Settings.statsFP);
    }
    if (Settings.statsToSyslog) {
	/* Chop off the date, because it's redundant.  The date is
	   always "dd/mm/YYYY:hh:mm:ss " = 20 characters long. */
	syslog(LOG_INFO, "stats %s", statbuf + 20);
    }
}

/**********************************************************************
* %FUNCTION: newGeneration
* %ARGUMENTS:
*  None
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Kills all running-but-idle slaves and increments Generation.  This
*  causes running-but-busy slaves to terminate at the earliest opportunity.
*  Use this if you change the filter file and want to restart the slaves.
***********************************************************************/
static void
newGeneration(void)
{
    Generation++;

    while(Slaves[STATE_IDLE]) {
	killSlave(Slaves[STATE_IDLE],
		  "Forcing reread of filter rules");
    }
}

/**********************************************************************
* %FUNCTION: bringSlavesUpToMin
* %ARGUMENTS:
*  es -- event selector
*  fd, flags, data -- ignored
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  If there are fewer than minSlaves running, start one.  If still fewer,
*  schedule self to re-run in slewTime seconds.
***********************************************************************/
void
bringSlavesUpToMin(EventSelector *es,
		   int fd,
		   unsigned int flags,
		   void *data)
{
    Slave *s;
    char reason[200];

    minScheduled = 0;

    if (NUM_RUNNING_SLAVES >= Settings.minSlaves) {
	/* Enough slaves, so do nothing */
	return;
    }

    /* Start a slave */
    s = Slaves[STATE_STOPPED];
    if (s) {
	snprintf(reason, sizeof(reason),
		 "Bringing slaves up to minSlaves (%d)", Settings.minSlaves);
	activateSlave(s, reason);
    }

    /* Reschedule if necessary */
    if (NUM_RUNNING_SLAVES < Settings.minSlaves) {
	scheduleBringSlavesUpToMin(es);
    }
}

/**********************************************************************
* %FUNCTION: scheduleBringSlavesUpToMin
* %ARGUMENTS:
*  es -- event selector
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Schedules a call to bringSlavesUpToMin in slewTime seconds from now.
***********************************************************************/
static void
scheduleBringSlavesUpToMin(EventSelector *es)
{
    struct timeval t;

    /* Do nothing if already scheduled */
    if (minScheduled) return;

    minScheduled = 1;
    t.tv_usec = 0;
    t.tv_sec = Settings.slewTime;
    Event_AddTimerHandler(es, t, bringSlavesUpToMin, NULL);
}

#ifdef HAVE_SETRLIMIT
/**********************************************************************
* %FUNCTION: limit_mem_usage
* %ARGUMENTS:
*  rss -- maximum resident-set size in kB
*  as  -- maximum total address space in kB
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Calls setrlimit to limit resource usage
***********************************************************************/
static void
limit_mem_usage(unsigned long rss,
		unsigned long as)
{
    int n;
    struct rlimit lim;

#ifdef RLIMIT_RSS
    if (rss) {
	/* Convert kb to bytes */
	rss *= 1024;
	lim.rlim_cur = rss;
	lim.rlim_max = rss;
	n = setrlimit(RLIMIT_RSS, &lim);
	if (n < 0) {
	    syslog(LOG_WARNING, "setrlimit(RLIMIT_RSS, %lu) failed: %m",
		   rss);
	}
    }
#endif
    if (as) {
	/* Convert kb to bytes */
	as *= 1024;
	lim.rlim_cur = as;
	lim.rlim_max = as;
#ifdef RLIMIT_AS
	n = setrlimit(RLIMIT_AS, &lim);
	if (n < 0) {
	    syslog(LOG_WARNING, "setrlimit(RLIMIT_AS, %lu) failed: %m",
		   as);
	}
#endif

#ifdef RLIMIT_DATA
	n = setrlimit(RLIMIT_DATA, &lim);
	if (n < 0) {
	    syslog(LOG_WARNING, "setrlimit(RLIMIT_DATA, %lu) failed: %m",
		   as);
	}
#endif

    }
}
#endif /* HAVE_SETRLIMIT */

/**********************************************************************
* %FUNCTION: doStatus
* %ARGUMENTS:
*  es -- event selector
*  fd -- socket
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Prints a status string back.  Each char corresponds to a slave;
*  the chars are:
*  S -- slave is stopped
*  I -- slave is idle
*  B -- slave is busy
*  K -- slave is killed but not yet reaped.
***********************************************************************/
static void
doStatus(EventSelector *es, int fd)
{
    char *ans = malloc(Settings.maxSlaves + 2 + 300);
    int i;

    if (!ans) {
	reply_to_mimedefang(es, fd, "error: Out of memory\n");
	return;
    }

    for (i=0; i < Settings.maxSlaves; i++) {
	Slave *s = &AllSlaves[i];
	switch (s->state) {
	case STATE_STOPPED: ans[i] = 'S'; break;
	case STATE_IDLE:    ans[i] = 'I'; break;
	case STATE_BUSY:    ans[i] = 'B'; break;
	case STATE_KILLED:  ans[i] = 'K'; break;
	default:            ans[i] = '?';
	}
    }
    sprintf(ans + Settings.maxSlaves, " %d %d\n", NumMsgsProcessed, Activations);
    reply_to_mimedefang(es, fd, ans);
}


/**********************************************************************
* %FUNCTION: findFreeSlave
* %ARGUMENTS:
*  None
* %RETURNS:
*  A pointer to a free slave, or NULL if none found.  Prefers to
*  return a running slave rather than one which needs activation.  Also,
*  prefers to return the slave which has been running the longest since
*  activation.
* %DESCRIPTION:
*  Finds a free (preferably running) slave.
***********************************************************************/
static Slave *
findFreeSlave(void)
{
    Slave *s = Slaves[STATE_IDLE];
    Slave *best = s;
    while(s) {
	if (s->activated < best->activated) {
	    best = s;
	}
	s = s->next;
    }

    if (best) return best;

    /* No running slaves - just pick the first stopped slave */
    return Slaves[STATE_STOPPED];
}

/**********************************************************************
* %FUNCTION: log_slave_resource_usage
* %ARGUMENTS:
*  s -- slave
*  usage -- struct rusage with slave's resource usage.
* %RETURNS:
*  Nothing
* %DESCRIPTION:
*  Logs slave's resource usage.
***********************************************************************/
#ifdef HAVE_WAIT3
static void
log_slave_resource_usage(Slave *s,
			 struct rusage *usage)
{
    if (!DOLOG) return;
    syslog(LOG_INFO,
	   "Slave %d resource usage: req=%d, scans=%d, user=%d.%03d, "
	   "sys=%d.%03d, nswap=%ld, majflt=%ld, minflt=%ld, "
	   "maxrss=%ld, bi=%ld, bo=%ld",
	   SLAVENO(s),
	   s->numRequests,
	   s->numScans,
	   (int) usage->ru_utime.tv_sec,
	   (int) usage->ru_utime.tv_usec / 1000,
	   (int) usage->ru_stime.tv_sec,
	   (int) usage->ru_stime.tv_usec / 1000,
	   usage->ru_nswap,
	   usage->ru_majflt,
	   usage->ru_minflt,
	   usage->ru_maxrss,
	   usage->ru_inblock,
	   usage->ru_oublock);
}
#endif
